import unittest

from rdflib import Literal
from rdflib import RDF
from rdflib import RDFS
from rdflib import URIRef
from rdflib import plugin
from rdflib import query
from rdflib.graph import ConjunctiveGraph
from rdflib.graph import Graph
from rdflib.graph import ReadOnlyGraphAggregate
from rdflib.store import Store
from six.moves import cStringIO as StringIO


plugin.register(
    'xml', query.ResultParser,
    'rdflib.plugins.sparql.results.xmlresults', 'XMLResultParser')
plugin.register(
    'xml', query.ResultSerializer,
    'rdflib.plugins.sparql.results.xmlresults', 'XMLResultSerializer')

plugin.register(
    'json', query.ResultParser,
    'rdflib.plugins.sparql.results.jsonresults', 'JSONResultParser')
plugin.register(
    'json', query.ResultSerializer,
    'rdflib.plugins.sparql.results.jsonresults', 'JSONResultSerializer')


testGraph1N3 = """
@prefix rdf:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix : <http://test/> .
:foo a rdfs:Class.
:bar :d :c.
:a :d :c.
"""

testGraph2N3 = """
@prefix rdf:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix : <http://test/> .
@prefix log: <http://www.w3.org/2000/10/swap/log#>.
:foo a rdfs:Resource.
:bar rdfs:isDefinedBy [ a log:Formula ].
:a :d :e.
"""

testGraph3N3 = """
@prefix rdf:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix log: <http://www.w3.org/2000/10/swap/log#>.
@prefix : <http://test/> .
<> a log:N3Document.
"""

sparqlQ3 =\
    """
PREFIX log: <http://www.w3.org/2000/10/swap/log#>
SELECT ?n3Doc
WHERE {?n3Doc a log:N3Document }"""


class GraphAggregates1(unittest.TestCase):
    dburi = Literal('sqlite://')

    def setUp(self):
        memStore = plugin.get('SQLAlchemy', Store)(
            identifier="rdflib_test", configuration=self.dburi)
        self.graph1 = Graph(memStore)
        self.graph2 = Graph(memStore)
        self.graph3 = Graph(memStore)

        for n3Str, graph in [(testGraph1N3, self.graph1),
                             (testGraph2N3, self.graph2),
                             (testGraph3N3, self.graph3)]:
            graph.parse(StringIO(n3Str), format='n3')

        self.G = ReadOnlyGraphAggregate(
            [self.graph1, self.graph2, self.graph3])

    def testAggregateRaw(self):
        # Test triples
        assert len(list(
            self.G.triples((None, RDF.type, None)))) == 4
        assert len(list(
            self.G.triples((URIRef("http://test/bar"), None, None)))) == 2
        assert len(list(
            self.G.triples((None, URIRef("http://test/d"), None)))) == 3

        # Test __len__
        # assert len(self.G) == 8, self.G.serialize(format="nt")
        assert len(list(self.G.triples((None, None, None)))) == 8

        # assert context iteration
        for g in self.G.contexts():
            assert isinstance(g, Graph)

        # Test __contains__
        assert (URIRef("http://test/foo"), RDF.type, RDFS.Resource) in self.G

        barPredicates = [URIRef("http://test/d"), RDFS.isDefinedBy]
        assert len(list(self.G.triples_choices(
            (URIRef("http://test/bar"), barPredicates, None)))) == 2


class GraphAggregates3(unittest.TestCase):
    def setUp(self):
        memStore = plugin.get('SQLAlchemy', Store)(
            identifier="rdflib_test", configuration=Literal("sqlite://"))
        self.graph1 = Graph(memStore, URIRef("graph1"))
        self.graph2 = Graph(memStore, URIRef("graph2"))
        self.graph3 = Graph(memStore, URIRef("graph3"))

        for n3Str, graph in [(testGraph1N3, self.graph1),
                             (testGraph2N3, self.graph2),
                             (testGraph3N3, self.graph3)]:
            graph.parse(StringIO(n3Str), format='n3')
        self.G = ConjunctiveGraph(memStore)

    def testDefaultGraph(self):
        # test that CG includes triples from all 3
        assert self.G.query(sparqlQ3), "CG as default graph should *all* triples"
        assert not self.graph2.query(sparqlQ3), "Graph as " + \
            "default graph should *not* include triples from other graphs"


if __name__ == '__main__':
    unittest.main()



"""Convenience functions for working with Terms and Graphs."""
from rdflib import BNode, Graph, Literal, URIRef, Variable
from rdflib.graph import QuotedGraph
from rdflib.term import Statement

from rdflib_sqlalchemy.constants import (
    TERM_COMBINATIONS,
    TERM_INSTANTIATION_DICT,
    REVERSE_TERM_COMBINATIONS,
)

__all__ = ["extract_triple"]


SUBJECT = 0
PREDICATE = 1
OBJECT = 2
CONTEXT = 3

GRAPH_TERM_DICT = {
    "F": (QuotedGraph, URIRef),
    "U": (Graph, URIRef),
    "B": (Graph, BNode)
}


def normalize_graph(graph):
    """
    Take an instance of a ``Graph`` and return the instance's identifier and  ``type``.

    Types are ``U`` for a :class:`~rdflib.graph.Graph`, ``F`` for
    a :class:`~rdflib.graph.QuotedGraph` and ``B`` for a
    :class:`~rdflib.graph.ConjunctiveGraph`

    >>> from rdflib import plugin
    >>> from rdflib.graph import Graph, ConjunctiveGraph, QuotedGraph
    >>> from rdflib.store import Store
    >>> from rdflib import URIRef, Namespace
    >>> from rdflib_sqlalchemy.termutils import normalize_graph
    >>> memstore = plugin.get('IOMemory', Store)()
    >>> g = Graph(memstore, URIRef("http://purl.org/net/bel-epa/gjh"))
    >>> normalize_graph(g)
    (rdflib.term.URIRef(u'http://purl.org/net/bel-epa/gjh'), 'U')
    >>> g = ConjunctiveGraph(memstore, Namespace("http://rdflib.net/ns"))
    >>> normalize_graph(g)  #doctest: +ELLIPSIS
    (rdflib.term.URIRef(u'http://rdflib.net/ns'), 'U')
    >>> g = QuotedGraph(memstore, Namespace("http://rdflib.net/ns"))
    >>> normalize_graph(g)
    (rdflib.term.URIRef(u'http://rdflib.net/ns'), 'F')

    """
    if isinstance(graph, QuotedGraph):
        return graph.identifier, "F"
    else:
        return graph.identifier, term_to_letter(graph.identifier)


def term_to_letter(term):
    """
    Relate a given term to one of several key types.

    * :class:`~rdflib.term.BNode`,
    * :class:`~rdflib.term.Literal`,
    * :class:`~rdflib.term.Statement` (Deprecated)
    * :class:`~rdflib.term.URIRef`,
    * :class:`~rdflib.term.Variable`
    * :class:`~rdflib.graph.Graph`
    * :class:`~rdflib.graph.QuotedGraph`

    >>> from rdflib import URIRef
    >>> from rdflib.term import BNode
    >>> # from rdflib.term import Statement
    >>> from rdflib.graph import Graph, QuotedGraph
    >>> from rdflib_sqlalchemy.termutils import term_to_letter
    >>> term_to_letter(URIRef('http://purl.org/net/bel-epa.com/'))
    'U'
    >>> term_to_letter(BNode())
    'B'
    >>> term_to_letter(Literal(u''))  # noqa
    'L'
    >>> term_to_letter(Variable(u'x'))  # noqa
    'V'
    >>> term_to_letter(Graph())
    'B'
    >>> term_to_letter(QuotedGraph("IOMemory", None))
    'F'
    >>> term_to_letter(None)
    'L'
    >>> # term_to_letter(Statement((None, None, None), None)) # Deprecated

    """
    if isinstance(term, URIRef):
        return "U"
    elif isinstance(term, BNode):
        return "B"
    elif isinstance(term, Literal):
        return "L"
    elif isinstance(term, QuotedGraph):
        return "F"
    elif isinstance(term, Variable):
        return "V"
    elif isinstance(term, Statement):
        return "s"
    elif isinstance(term, Graph):
        return term_to_letter(term.identifier)
    elif term is None:
        return "L"
    else:
        raise Exception(
            ("The given term (%s) is not an instance of any " +
             "of the known types (URIRef, BNode, Literal, QuotedGraph, " +
             "or Variable).  It is a %s")
            % (term, type(term)))


def construct_graph(key):
    """
    Return a tuple containing a ``Graph`` and an appropriate referent.

    Takes a key (one of 'F', 'U' or 'B')

    >>> from rdflib_sqlalchemy.termutils import construct_graph
    >>> construct_graph('F')
    (<class 'rdflib.graph.QuotedGraph'>, <class 'rdflib.term.URIRef'>)
    >>> construct_graph('U')
    (<class 'rdflib.graph.Graph'>, <class 'rdflib.term.URIRef'>)
    >>> construct_graph('B')
    (<class 'rdflib.graph.Graph'>, <class 'rdflib.term.BNode'>)

    """
    return GRAPH_TERM_DICT[key]


def triple_pattern_to_term_combinations(triple):
    """Map a triple pattern to term combinations (non-functioning)."""
    s, p, o = triple
    combinations = []
    if isinstance(o, Literal):
        for key, val in TERM_COMBINATIONS.items():
            if key[OBJECT] == 'O':
                combinations.append(val)
    return combinations


def type_to_term_combination(member, klass, context):
    """Map a type to a term combination."""
    term_combination = '{subject}U{object}{context}'.format(
        subject=term_to_letter(member),
        object=term_to_letter(klass),
        context=normalize_graph(context)[-1],
    )

    try:
        return TERM_COMBINATIONS[term_combination]
    except KeyError:
        if isinstance(member, Literal):
            raise ValueError(
                'A Literal cannot be a subject of a triple.\n\n'
                'Triple causing error:\n'
                '  {member} rdf:type {klass}\n'
                'Context: {context}'.format(
                    member=member,
                    klass=klass,
                    context=context,
                )
            )

        raise


def statement_to_term_combination(subject, predicate, obj, context):
    """Map a statement to a Term Combo."""
    return TERM_COMBINATIONS["%s%s%s%s" %
                             (term_to_letter(subject), term_to_letter(predicate),
                              term_to_letter(obj), normalize_graph(context)[-1])]


def escape_quotes(qstr):
    """
    Escape backslashes.

    #FIXME:  This *may* prove to be a performance bottleneck and should
             perhaps be implemented in C (as it was in 4Suite RDF)

    Ported from Ft.Lib.DbUtil
    """
    if qstr is None:
        return ""
    tmp = qstr.replace("\\", "\\\\")
    tmp = tmp.replace("'", "\\'")
    return tmp


def extract_triple(tupleRt, store, hardCodedContext=None):
    """
    Extract a triple.

    Take a tuple which represents an entry in a result set and
    converts it to a tuple of terms using the termComb integer
    to interpret how to instantiate each term.

    """
    try:
        id, subject, predicate, obj, rtContext, termComb, \
            objLanguage, objDatatype = tupleRt
        termCombString = REVERSE_TERM_COMBINATIONS[termComb]
        subjTerm, predTerm, objTerm, ctxTerm = termCombString
    except ValueError:
        id, subject, subjTerm, predicate, predTerm, obj, objTerm, \
            rtContext, ctxTerm, objLanguage, objDatatype = tupleRt

    context = rtContext is not None \
        and rtContext \
        or hardCodedContext.identifier
    s = create_term(subject, subjTerm, store)
    p = create_term(predicate, predTerm, store)
    o = create_term(obj, objTerm, store, objLanguage, objDatatype)

    graphKlass, idKlass = construct_graph(ctxTerm)

    return id, s, p, o, (graphKlass, idKlass, context)


def create_term(termString, termType, store, objLanguage=None, objDatatype=None):
    """
    Take a term value, term type, and store instance and creates a term object.

    QuotedGraphs are instantiated differently
    """
    if termType == "L":
        cache = store.literalCache.get((termString, objLanguage, objDatatype))
        if cache is not None:
            # store.cacheHits += 1
            return cache
        else:
            # store.cacheMisses += 1
            # rt = Literal(termString, objLanguage, objDatatype)
            # store.literalCache[((termString, objLanguage, objDatatype))] = rt
            if objLanguage and not objDatatype:
                rt = Literal(termString, objLanguage)
                store.literalCache[((termString, objLanguage))] = rt
            elif objDatatype and not objLanguage:
                rt = Literal(termString, datatype=objDatatype)
                store.literalCache[((termString, objDatatype))] = rt
            elif not objLanguage and not objDatatype:
                rt = Literal(termString)
                store.literalCache[((termString))] = rt
            else:
                rt = Literal(termString, objDatatype)
                store.literalCache[((termString, objDatatype))] = rt
            return rt
    elif termType == "F":
        cache = store.otherCache.get((termType, termString))
        if cache is not None:
            # store.cacheHits += 1
            return cache
        else:
            # store.cacheMisses += 1
            rt = QuotedGraph(store, URIRef(termString))
            store.otherCache[(termType, termString)] = rt
            return rt
    elif termType == "B":
        cache = store.bnodeCache.get((termString))
        if cache is not None:
            # store.cacheHits += 1
            return cache
        else:
            # store.cacheMisses += 1
            rt = TERM_INSTANTIATION_DICT[termType](termString)
            store.bnodeCache[(termString)] = rt
            return rt
    elif termType == "U":
        cache = store.uriCache.get((termString))
        if cache is not None:
            # store.cacheHits += 1
            return cache
        else:
            # store.cacheMisses += 1
            rt = URIRef(termString)
            store.uriCache[(termString)] = rt
            return rt
    else:
        cache = store.otherCache.get((termType, termString))
        if cache is not None:
            # store.cacheHits += 1
            return cache
        else:
            # store.cacheMisses += 1
            rt = TERM_INSTANTIATION_DICT[termType](termString)
            store.otherCache[(termType, termString)] = rt
            return rt



from sqlalchemy import Column, Table, Index, types

from rdflib_sqlalchemy.types import TermType


MYSQL_MAX_INDEX_LENGTH = 200

TABLE_NAME_TEMPLATES = [
    "{interned_id}_asserted_statements",
    "{interned_id}_literal_statements",
    "{interned_id}_namespace_binds",
    "{interned_id}_quoted_statements",
    "{interned_id}_type_statements",
]


def get_table_names(interned_id):
    return [
        table_name_template.format(interned_id=interned_id)
        for table_name_template in TABLE_NAME_TEMPLATES
    ]


def create_asserted_statements_table(interned_id, metadata):
    return Table(
        "{interned_id}_asserted_statements".format(interned_id=interned_id),
        metadata,
        Column("id", types.Integer, nullable=False, primary_key=True),
        Column("subject", TermType, nullable=False),
        Column("predicate", TermType, nullable=False),
        Column("object", TermType, nullable=False),
        Column("context", TermType, nullable=False),
        Column("termcomb", types.Integer, nullable=False, key="termComb"),
        Index(
            "{interned_id}_A_s_index".format(interned_id=interned_id),
            "subject",
            mysql_length=MYSQL_MAX_INDEX_LENGTH,
        ),
        Index(
            "{interned_id}_A_p_index".format(interned_id=interned_id),
            "predicate",
            mysql_length=MYSQL_MAX_INDEX_LENGTH,
        ),
        Index(
            "{interned_id}_A_o_index".format(interned_id=interned_id),
            "object",
            mysql_length=MYSQL_MAX_INDEX_LENGTH,
        ),
        Index(
            "{interned_id}_A_c_index".format(interned_id=interned_id),
            "context",
            mysql_length=MYSQL_MAX_INDEX_LENGTH,
        ),
        Index(
            "{interned_id}_A_termComb_index".format(interned_id=interned_id),
            "termComb",
        ),
        Index(
            "{interned_id}_asserted_spoc_key".format(interned_id=interned_id),
            "subject",
            "predicate",
            "object",
            "context",
            unique=True,
            mysql_length=191,
        ),
    )


def create_type_statements_table(interned_id, metadata):
    return Table(
        "{interned_id}_type_statements".format(interned_id=interned_id),
        metadata,
        Column("id", types.Integer, nullable=False, primary_key=True),
        Column("member", TermType, nullable=False),
        Column("klass", TermType, nullable=False),
        Column("context", TermType, nullable=False),
        Column("termcomb", types.Integer, nullable=False, key="termComb"),
        Index(
            "{interned_id}_member_index".format(interned_id=interned_id),
            "member",
            mysql_length=MYSQL_MAX_INDEX_LENGTH,
        ),
        Index(
            "{interned_id}_klass_index".format(interned_id=interned_id),
            "klass",
            mysql_length=MYSQL_MAX_INDEX_LENGTH,
        ),
        Index(
            "{interned_id}_c_index".format(interned_id=interned_id),
            "context",
            mysql_length=MYSQL_MAX_INDEX_LENGTH,
        ),
        Index(
            "{interned_id}_T_termComb_index".format(interned_id=interned_id),
            "termComb",
        ),
        Index(
            "{interned_id}_type_mkc_key".format(interned_id=interned_id),
            "member",
            "klass",
            "context",
            unique=True,
            mysql_length=MYSQL_MAX_INDEX_LENGTH,
        ),
    )


def create_literal_statements_table(interned_id, metadata):
    return Table(
        "{interned_id}_literal_statements".format(interned_id=interned_id),
        metadata,
        Column("id", types.Integer, nullable=False, primary_key=True),
        Column("subject", TermType, nullable=False),
        Column("predicate", TermType, nullable=False),
        Column("object", TermType),
        Column("context", TermType, nullable=False),
        Column("termcomb", types.Integer, nullable=False, key="termComb"),
        Column("objlanguage", types.String(255), key="objLanguage"),
        Column("objdatatype", types.String(255), key="objDatatype"),
        Index(
            "{interned_id}_L_s_index".format(interned_id=interned_id),
            "subject",
            mysql_length=MYSQL_MAX_INDEX_LENGTH,
        ),
        Index(
            "{interned_id}_L_p_index".format(interned_id=interned_id),
            "predicate",
            mysql_length=MYSQL_MAX_INDEX_LENGTH,
        ),
        Index(
            "{interned_id}_L_c_index".format(interned_id=interned_id),
            "context",
            mysql_length=MYSQL_MAX_INDEX_LENGTH,
        ),
        Index(
            "{interned_id}_L_termComb_index".format(interned_id=interned_id),
            "termComb",
        ),
        Index(
            "{interned_id}_literal_spoc_key".format(interned_id=interned_id),
            "subject",
            "predicate",
            "object",
            "objLanguage",
            "context",
            unique=True,
            mysql_length=153,
        ),
    )


def create_quoted_statements_table(interned_id, metadata):
    return Table(
        "{interned_id}_quoted_statements".format(interned_id=interned_id),
        metadata,
        Column("id", types.Integer, nullable=False, primary_key=True),
        Column("subject", TermType, nullable=False),
        Column("predicate", TermType, nullable=False),
        Column("object", TermType),
        Column("context", TermType, nullable=False),
        Column("termcomb", types.Integer, nullable=False, key="termComb"),
        Column("objlanguage", types.String(255), key="objLanguage"),
        Column("objdatatype", types.String(255), key="objDatatype"),
        Index(
            "{interned_id}_Q_s_index".format(interned_id=interned_id),
            "subject",
            mysql_length=MYSQL_MAX_INDEX_LENGTH,
        ),
        Index(
            "{interned_id}_Q_p_index".format(interned_id=interned_id),
            "predicate",
            mysql_length=MYSQL_MAX_INDEX_LENGTH,
        ),
        Index(
            "{interned_id}_Q_o_index".format(interned_id=interned_id),
            "object",
            mysql_length=MYSQL_MAX_INDEX_LENGTH,
        ),
        Index(
            "{interned_id}_Q_c_index".format(interned_id=interned_id),
            "context",
            mysql_length=MYSQL_MAX_INDEX_LENGTH,
        ),
        Index(
            "{interned_id}_Q_termComb_index".format(interned_id=interned_id),
            "termComb",
        ),
        Index(
            "{interned_id}_quoted_spoc_key".format(interned_id=interned_id),
            "subject",
            "predicate",
            "object",
            "objLanguage",
            "context",
            unique=True,
            mysql_length=153,
        ),
    )


def create_namespace_binds_table(interned_id, metadata):
    return Table(
        "{interned_id}_namespace_binds".format(interned_id=interned_id),
        metadata,
        Column("prefix", types.String(20), unique=True, nullable=False, primary_key=True),
        Column("uri", types.Text),
        Index(
            "{interned_id}_uri_index".format(interned_id=interned_id),
            "uri",
            mysql_length=MYSQL_MAX_INDEX_LENGTH,
        )
    )



import logging
import os
import unittest

from nose import SkipTest
from rdflib import Literal
from rdflib.graph import ConjunctiveGraph as Graph
from six import PY3
from rdflib.store import NO_STORE, VALID_STORE
from rdflib.term import URIRef

from . import context_case
from . import graph_case


if PY3:
    try:
        import mysql
        assert mysql
        dialect = "mysqlconnector"
    except ImportError:
        raise SkipTest("MySQL-connector not found, skipping MySQL tests")
else:
    try:
        import MySQLdb
        assert MySQLdb
        dialect = "mysqldb"
    except ImportError:
        raise SkipTest("MySQLdb not found, skipping MySQL tests")


if os.environ.get("DB") != "mysql":
    raise SkipTest("MySQL not under test")

_logger = logging.getLogger(__name__)


sqlalchemy_url = Literal(os.environ.get(
    "DBURI",
    "mysql+%s://root@127.0.0.1:3306/test?charset=utf8" % dialect))


class SQLAMySQLGraphTestCase(graph_case.GraphTestCase):
    storetest = True
    storename = "SQLAlchemy"
    uri = sqlalchemy_url
    create = True

    def setUp(self):
        super(SQLAMySQLGraphTestCase, self).setUp(uri=self.uri, storename=self.storename)

    def tearDown(self):
        super(SQLAMySQLGraphTestCase, self).tearDown(uri=self.uri)


class SQLAMySQLContextTestCase(context_case.ContextTestCase):
    storetest = True
    storename = "SQLAlchemy"
    uri = sqlalchemy_url
    create = True

    def setUp(self):
        super(SQLAMySQLContextTestCase, self).setUp(
            uri=self.uri, storename=self.storename)

    def tearDown(self):
        super(SQLAMySQLContextTestCase, self).tearDown(uri=self.uri)

    def testLenInMultipleContexts(self):
        raise SkipTest("Known issue.")


class SQLAMySQLIssueTestCase(unittest.TestCase):
    storetest = True
    storename = "SQLAlchemy"
    uri = sqlalchemy_url

    def test_issue_4(self):
        ident = URIRef("rdflib_test")
        g = Graph(store="SQLAlchemy", identifier=ident)
        rt = g.open(self.uri, create=True)
        if rt == NO_STORE:
            g.open(self.uri, create=True)
        else:
            assert rt == VALID_STORE, "The underlying store is not valid: State: %s" % rt
        g.destroy(self.uri)


if __name__ == "__main__":
    unittest.main()



import logging
import os
import unittest

from nose import SkipTest
from rdflib import Literal

from . import context_case
from . import graph_case


if os.environ.get("DB") != "sqlite":
    raise SkipTest("SQLite not under test")

_logger = logging.getLogger(__name__)

sqlalchemy_url = Literal(os.environ.get("DBURI", "sqlite://"))


class SQLASQLiteGraphTestCase(graph_case.GraphTestCase):
    storetest = True
    storename = "SQLAlchemy"
    uri = sqlalchemy_url

    def setUp(self):
        super(SQLASQLiteGraphTestCase, self).setUp(
            uri=self.uri, storename=self.storename)

    def tearDown(self):
        super(SQLASQLiteGraphTestCase, self).tearDown(uri=self.uri)


class SQLASQLiteContextTestCase(context_case.ContextTestCase):
    storetest = True
    storename = "SQLAlchemy"
    uri = sqlalchemy_url

    def setUp(self):
        super(SQLASQLiteContextTestCase, self).setUp(
            uri=self.uri, storename=self.storename)

    def tearDown(self):
        super(SQLASQLiteContextTestCase, self).tearDown(uri=self.uri)

    def testLenInMultipleContexts(self):
        raise SkipTest("Known issue.")


SQLASQLiteGraphTestCase.storetest = True
SQLASQLiteContextTestCase.storetest = True


if __name__ == "__main__":
    unittest.main()



from rdflib.graph import Graph, QuotedGraph
from rdflib.term import Node
from six import text_type
from sqlalchemy import types


class TermType(types.TypeDecorator):
    """Term typology."""

    impl = types.Text()
    cache_ok = True

    def process_bind_param(self, value, dialect):
        """Process bound parameters."""
        if isinstance(value, (QuotedGraph, Graph)):
            return text_type(value.identifier)
        elif isinstance(value, Node):
            return text_type(value)
        else:
            return value



"""SQLAlchemy-based RDF store."""
import hashlib
import logging

import sqlalchemy
from rdflib import (
    BNode,
    Literal,
    URIRef
)
from rdflib.term import Statement, Variable
from rdflib.graph import Graph, QuotedGraph
from rdflib.namespace import RDF
from rdflib.plugins.stores.regexmatching import PYTHON_REGEX, REGEXTerm
from rdflib.store import CORRUPTED_STORE, VALID_STORE, NodePickler, Store
from six import text_type
from sqlalchemy import MetaData, inspect
from sqlalchemy.sql import expression, select
from sqlalchemy.exc import OperationalError

from rdflib_sqlalchemy.constants import (
    ASSERTED_LITERAL_PARTITION,
    ASSERTED_NON_TYPE_PARTITION,
    ASSERTED_TYPE_PARTITION,
    CONTEXT_SELECT,
    COUNT_SELECT,
    INTERNED_PREFIX,
    QUOTED_PARTITION,
    TRIPLE_SELECT_NO_ORDER,
)
from rdflib_sqlalchemy.tables import (
    create_asserted_statements_table,
    create_literal_statements_table,
    create_namespace_binds_table,
    create_quoted_statements_table,
    create_type_statements_table,
    get_table_names,
)
from rdflib_sqlalchemy.base import SQLGeneratorMixin
from rdflib_sqlalchemy.sql import union_select
from rdflib_sqlalchemy.statistics import StatisticsMixin
from rdflib_sqlalchemy.termutils import extract_triple


_logger = logging.getLogger(__name__)

Any = None


def grouper(iterable, n):
    "Collect data into chunks of at most n elements"
    assert n > 0, 'Cannot group into chunks of zero elements'
    lst = []
    iterable = iter(iterable)
    while True:
        try:
            lst.append(next(iterable))
        except StopIteration:
            break

        if len(lst) == n:
            yield lst
            lst = []

    if lst:
        yield lst


def generate_interned_id(identifier):
    return "{prefix}{identifier_hash}".format(
        prefix=INTERNED_PREFIX,
        identifier_hash=hashlib.sha1(identifier.encode("utf8")).hexdigest()[:10],
    )


class SQLAlchemy(Store, SQLGeneratorMixin, StatisticsMixin):
    """
    SQL-92 formula-aware implementation of an rdflib Store.

    It stores its triples in the following partitions:

    - Asserted non rdf:type statements
    - Asserted literal statements
    - Asserted rdf:type statements (in a table which models Class membership)
        The motivation for this partition is primarily query speed and
        scalability as most graphs will always have more rdf:type statements
        than others
    - All Quoted statements

    In addition it persists namespace mappings in a separate table
    """

    context_aware = True
    formula_aware = True
    transaction_aware = True
    regex_matching = PYTHON_REGEX
    configuration = Literal("sqlite://")

    def __init__(self, identifier=None, configuration=None, engine=None,
                 max_terms_per_where=800):
        """
        Initialisation.

        Args:
            identifier (rdflib.URIRef): URIRef of the Store. Defaults to CWD.
            configuration: the database connection URL string or a configuration dictionary
                corresponding to the connection options accepted by sqlalchemy.create_engine,
                with the additional "url" key pointing to the connection URL. See `open` documentation
                for more details.
            engine (sqlalchemy.engine.Engine, optional): a pre-existing `SQLAlchemy.engine.Engine` instance.
            max_terms_per_where (int): The max number of terms (s/p/o) in a call to
                triples_choices to combine in one SQL "where" clause. Important for SQLite
                back-end with SQLITE_MAX_EXPR_DEPTH limit and SQLITE_LIMIT_COMPOUND_SELECT
                -- must find a balance that doesn't hit either of those.
        """
        self.identifier = identifier and identifier or "hardcoded"
        self.engine = engine
        self.max_terms_per_where = max_terms_per_where

        # Use only the first 10 bytes of the digest
        self._interned_id = generate_interned_id(self.identifier)

        # This parameter controls how exclusively the literal table is searched
        # If true, the Literal partition is searched *exclusively* if the
        # object term in a triple pattern is a Literal or a REGEXTerm.  Note,
        # the latter case prevents the matching of URIRef nodes as the objects
        # of a triple in the store.
        # If the object term is a wildcard (None)
        # Then the Literal partition is searched in addition to the others
        # If this parameter is false, the literal partition is searched
        # regardless of what the object of the triple pattern is
        self.STRONGLY_TYPED_TERMS = False

        self.cacheHits = 0
        self.cacheMisses = 0
        self.literalCache = {}
        self.uriCache = {}
        self.bnodeCache = {}
        self.otherCache = {}
        self._node_pickler = None

        self._create_table_definitions()

        # XXX For backward compatibility we still support getting the connection string in constructor
        # TODO: deprecate this once refactoring is more mature
        super(SQLAlchemy, self).__init__(configuration)

    def __repr__(self):
        """Readable serialisation."""
        quoted_table = self.tables["quoted_statements"]
        asserted_table = self.tables["asserted_statements"]
        asserted_type_table = self.tables["type_statements"]
        literal_table = self.tables["literal_statements"]

        selects = [
            (expression.alias(asserted_type_table, "typetable"), None, ASSERTED_TYPE_PARTITION),
            (expression.alias(quoted_table, "quoted"), None, QUOTED_PARTITION),
            (expression.alias(asserted_table, "asserted"), None, ASSERTED_NON_TYPE_PARTITION),
            (expression.alias(literal_table, "literal"), None, ASSERTED_LITERAL_PARTITION),
        ]
        q = union_select(selects, distinct=False, select_type=COUNT_SELECT)
        if hasattr(self, "engine"):
            with self.engine.connect() as connection:
                res = connection.execute(q)
                rt = res.fetchall()
                typeLen, quotedLen, assertedLen, literalLen = [
                    rtTuple[0] for rtTuple in rt]
            try:
                return ("<Partitioned SQL N3 Store: %s "
                        "contexts, %s classification assertions, "
                        "%s quoted statements, %s property/value "
                        "assertions, and %s other assertions>" % (
                            sum(1 for _ in self.contexts()),
                            typeLen, quotedLen, literalLen, assertedLen))
            except Exception:
                _logger.exception('Error creating repr')
                return "<Partitioned SQL N3 Store>"
        else:
            return "<Partitioned unopened SQL N3 Store>"

    def __len__(self, context=None):
        """Number of statements in the store."""
        quoted_table = self.tables["quoted_statements"]
        asserted_table = self.tables["asserted_statements"]
        asserted_type_table = self.tables["type_statements"]
        literal_table = self.tables["literal_statements"]

        typetable = expression.alias(asserted_type_table, "typetable")
        quoted = expression.alias(quoted_table, "quoted")
        asserted = expression.alias(asserted_table, "asserted")
        literal = expression.alias(literal_table, "literal")

        quotedContext = self.build_context_clause(context, quoted)
        assertedContext = self.build_context_clause(context, asserted)
        typeContext = self.build_context_clause(context, typetable)
        literalContext = self.build_context_clause(context, literal)

        if context is not None:
            selects = [
                (typetable, typeContext,
                 ASSERTED_TYPE_PARTITION),
                (quoted, quotedContext,
                 QUOTED_PARTITION),
                (asserted, assertedContext,
                 ASSERTED_NON_TYPE_PARTITION),
                (literal, literalContext,
                 ASSERTED_LITERAL_PARTITION), ]
            q = union_select(selects, distinct=True, select_type=COUNT_SELECT)
        else:
            selects = [
                (typetable, typeContext,
                 ASSERTED_TYPE_PARTITION),
                (asserted, assertedContext,
                 ASSERTED_NON_TYPE_PARTITION),
                (literal, literalContext,
                 ASSERTED_LITERAL_PARTITION), ]
            q = union_select(selects, distinct=False, select_type=COUNT_SELECT)

        with self.engine.connect() as connection:
            res = connection.execute(q)
            rt = res.fetchall()
            return int(sum(rtTuple[0] for rtTuple in rt))

    @property
    def table_names(self):
        return get_table_names(interned_id=self._interned_id)

    @property
    def node_pickler(self):
        if getattr(self, "_node_pickler", False) or self._node_pickler is None:
            self._node_pickler = np = NodePickler()
            np.register(self, "S")
            np.register(URIRef, "U")
            np.register(BNode, "B")
            np.register(Literal, "L")
            np.register(Graph, "G")
            np.register(QuotedGraph, "Q")
            np.register(Variable, "V")
            np.register(Statement, "s")
        return self._node_pickler

    def open(self, configuration, create=True):
        """
        Open the store specified by the configuration parameter.

        Args:
            configuration: if a string, use as the DBAPI URL. If a dictionary, will use as the
                `**kwargs` for the sqlalchemy.create_engine() call, and will attempt to
                extract the connection URL from a 'url' key in that dictionary.
                A valid connection string will be of the format:

                    dialect[+driver]://user:password@host/dbname[?key=value..]

            create (bool): If create is True a store will be created if it does not already
                exist. If create is False and a store does not already exist
                an exception is raised. An exception is also raised if a store
                exists, but there is insufficient permissions to open the
                store.

        Returns:
            int:
            - CORRUPTED_STORE (0) if database exists but is empty,
            - VALID_STORE (1) if database exists and tables are all there,
            - NO_STORE (-1) if nothing exists

        """
        # Close any existing engine connection
        self.close()

        url, kwargs = configuration, {}
        if isinstance(configuration, dict):
            url = configuration.pop("url", None)
            if not url:
                raise Exception('Configuration dict is missing the required "url" key')
            kwargs = configuration

        self.engine = sqlalchemy.create_engine(url, **kwargs)
        try:
            conn = self.engine.connect()
        except OperationalError:
            raise RuntimeError("open() - failed during engine connection")
        else:
            with conn:
                if create:
                    self.create_all()

                ret_value = self._verify_store_exists()

        if ret_value != VALID_STORE and not create:
            raise RuntimeError("open() - create flag was set to False, but store was not created previously.")

        return ret_value

    def create_all(self):
        """Create all of the database tables (idempotent)."""
        self.metadata.create_all(self.engine)

    def close(self, commit_pending_transaction=False):
        """
        Close the current store engine connection if one is open.
        """
        if self.engine:
            self.engine.dispose()
        self.engine = None

    def destroy(self, configuration):
        """
        Delete all tables and stored data associated with the store.
        """
        if self.engine is None:
            self.engine = self.open(configuration, create=False)

        with self.engine.begin():
            try:
                self.metadata.drop_all(self.engine)
            except Exception:
                _logger.exception("unable to drop table.")
                raise

    # Triple Methods

    def add(self, triple, context=None, quoted=False):
        """Add a triple to the store of triples."""
        super(SQLAlchemy, self).add(triple, context, quoted)
        subject, predicate, obj = triple
        _, statement, params = self._get_build_command(
            (subject, predicate, obj),
            context, quoted,
        )

        statement = self._add_ignore_on_conflict(statement)
        with self.engine.begin() as connection:
            try:
                connection.execute(statement, params)
            except Exception:
                _logger.exception(
                    "Add failed with statement: %s, params: %s",
                    str(statement), repr(params)
                )
                raise

    def addN(self, quads):
        """Add a list of triples in quads form."""
        commands_dict = {}
        add_event = super(SQLAlchemy, self).add
        for subject, predicate, obj, context in quads:
            add_event((subject, predicate, obj), context)
            command_type, statement, params = self._get_build_command(
                (subject, predicate, obj),
                context,
                isinstance(context, QuotedGraph),
            )

            command_dict = commands_dict.setdefault(command_type, {})
            command_dict.setdefault("statement", statement)
            command_dict.setdefault("params", []).append(params)

        with self.engine.begin() as connection:
            try:
                for command in commands_dict.values():
                    statement = self._add_ignore_on_conflict(command['statement'])
                    connection.execute(statement, command["params"])
            except Exception:
                _logger.exception("AddN failed.")
                raise

    def _add_ignore_on_conflict(self, statement):
        if self.engine.name == 'sqlite':
            statement = statement.prefix_with('OR IGNORE')
        elif self.engine.name == 'mysql':
            statement = statement.prefix_with('IGNORE')
        elif self.engine.name == 'postgresql':
            from sqlalchemy.dialects.postgresql.dml import OnConflictDoNothing
            statement._post_values_clause = OnConflictDoNothing()
        return statement

    def remove(self, triple, context):
        """Remove a triple from the store."""
        super(SQLAlchemy, self).remove(triple, context)
        subject, predicate, obj = triple

        if context is not None:
            if subject is None and predicate is None and object is None:
                self._remove_context(context)
                return

        quoted_table = self.tables["quoted_statements"]
        asserted_table = self.tables["asserted_statements"]
        asserted_type_table = self.tables["type_statements"]
        literal_table = self.tables["literal_statements"]

        with self.engine.begin() as connection:
            try:
                if predicate is None or predicate != RDF.type:
                    # Need to remove predicates other than rdf:type

                    if not self.STRONGLY_TYPED_TERMS or isinstance(obj, Literal):
                        # remove literal triple
                        clause = self.build_clause(literal_table, subject, predicate, obj, context)
                        connection.execute(literal_table.delete(clause))

                    for table in [quoted_table, asserted_table]:
                        # If asserted non rdf:type table and obj is Literal,
                        # don't do anything (already taken care of)
                        if table == asserted_table and isinstance(obj, Literal):
                            continue
                        else:
                            clause = self.build_clause(table, subject, predicate, obj, context)
                            connection.execute(table.delete(clause))

                if predicate == RDF.type or predicate is None:
                    # Need to check rdf:type and quoted partitions (in addition
                    # perhaps)
                    clause = self.build_clause(asserted_type_table, subject, RDF.type, obj, context, True)
                    connection.execute(asserted_type_table.delete(clause))

                    clause = self.build_clause(quoted_table, subject, predicate, obj, context)
                    connection.execute(quoted_table.delete(clause))
            except Exception:
                _logger.exception("Removal failed.")
                raise

    def _triples_helper(self, triple, context=None):
        subject, predicate, obj = triple

        quoted_table = self.tables["quoted_statements"]
        asserted_table = self.tables["asserted_statements"]
        asserted_type_table = self.tables["type_statements"]
        literal_table = self.tables["literal_statements"]

        if predicate == RDF.type:
            # select from asserted rdf:type partition and quoted table
            # (if a context is specified)
            typeTable = expression.alias(
                asserted_type_table, "typetable")
            clause = self.build_clause(typeTable, subject, RDF.type, obj, context, True)
            selects = [
                (typeTable,
                 clause,
                 ASSERTED_TYPE_PARTITION), ]

        elif isinstance(predicate, REGEXTerm) \
                and predicate.compiledExpr.match(RDF.type) \
                or predicate is None:
            # Select from quoted partition (if context is specified),
            # Literal partition if (obj is Literal or None) and asserted
            # non rdf:type partition (if obj is URIRef or None)
            selects = []
            if (not self.STRONGLY_TYPED_TERMS
                    or isinstance(obj, Literal)
                    or obj is None
                    or (self.STRONGLY_TYPED_TERMS and isinstance(obj, REGEXTerm))):
                literal = expression.alias(literal_table, "literal")
                clause = self.build_clause(literal, subject, predicate, obj, context)
                selects.append((literal, clause, ASSERTED_LITERAL_PARTITION))

            if not isinstance(obj, Literal) \
                    and not (isinstance(obj, REGEXTerm) and self.STRONGLY_TYPED_TERMS) \
                    or obj is None:
                asserted = expression.alias(asserted_table, "asserted")
                clause = self.build_clause(asserted, subject, predicate, obj, context)
                selects.append((asserted, clause, ASSERTED_NON_TYPE_PARTITION))

            typeTable = expression.alias(asserted_type_table, "typetable")
            clause = self.build_clause(typeTable, subject, RDF.type, obj, context, True)
            selects.append((typeTable, clause, ASSERTED_TYPE_PARTITION))

        elif predicate:
            # select from asserted non rdf:type partition (optionally),
            # quoted partition (if context is specified), and literal
            # partition (optionally)
            selects = []
            if not self.STRONGLY_TYPED_TERMS \
                    or isinstance(obj, Literal) \
                    or obj is None \
                    or (self.STRONGLY_TYPED_TERMS and isinstance(obj, REGEXTerm)):
                literal = expression.alias(literal_table, "literal")
                clause = self.build_clause(literal, subject, predicate, obj, context)
                selects.append((literal, clause, ASSERTED_LITERAL_PARTITION))

            if (obj is None or (not isinstance(obj, Literal)
                    and not (isinstance(obj, REGEXTerm) and self.STRONGLY_TYPED_TERMS))):
                asserted = expression.alias(asserted_table, "asserted")
                clause = self.build_clause(asserted, subject, predicate, obj, context)
                selects.append((asserted, clause, ASSERTED_NON_TYPE_PARTITION))

        if context is not None:
            quoted = expression.alias(quoted_table, "quoted")
            clause = self.build_clause(quoted, subject, predicate, obj, context)
            selects.append((quoted, clause, QUOTED_PARTITION))

        return selects

    def _do_triples_select(self, selects, context):
        q = union_select(selects, distinct=True, select_type=TRIPLE_SELECT_NO_ORDER)
        with self.engine.connect() as connection:
            res = connection.execute(q)
            # TODO: False but it may have limitations on text column. Check
            # NOTE: SQLite does not support ORDER BY terms that aren't
            # integers, so the entire result set must be iterated in order
            # to be able to return a generator of contexts
            result = res.fetchall()
        tripleCoverage = {}

        for rt in result:
            id, s, p, o, (graphKlass, idKlass, graphId) = extract_triple(rt, self, context)
            contexts = tripleCoverage.get((s, p, o), [])
            contexts.append(graphKlass(self, idKlass(graphId)))
            tripleCoverage[(s, p, o)] = contexts

        for (s, p, o), contexts in tripleCoverage.items():
            yield (s, p, o), (c for c in contexts)

    def triples(self, triple, context=None):
        """ A generator over all the triples matching a pattern. """
        selects = self._triples_helper(triple, context)
        for m in self._do_triples_select(selects, context):
            yield m

    def triples_choices(self, triple, context=None):
        """
        A variant of triples.
        """
        # We already support accepting a list for s/p/o
        subject, predicate, object_ = triple
        selects = []
        if isinstance(object_, list):
            assert not isinstance(
                subject, list), "object_ / subject are both lists"
            assert not isinstance(
                predicate, list), "object_ / predicate are both lists"
            if not object_:
                object_ = None
            for o in grouper(object_, self.max_terms_per_where):
                for sels in self._triples_helper((subject, predicate, o), context):
                    selects.append(sels)

        elif isinstance(subject, list):
            assert not isinstance(
                predicate, list), "subject / predicate are both lists"
            if not subject:
                subject = None
            for s in grouper(subject, self.max_terms_per_where):
                for sels in self._triples_helper((s, predicate, object_), context):
                    selects.append(sels)

        elif isinstance(predicate, list):
            assert not isinstance(
                subject, list), "predicate / subject are both lists"
            if not predicate:
                predicate = None
            for p in grouper(predicate, self.max_terms_per_where):
                for sels in self._triples_helper((subject, p, object_), context):
                    selects.append(sels)

        for m in self._do_triples_select(selects, context):
            yield m

    def contexts(self, triple=None):
        quoted_table = self.tables["quoted_statements"]
        asserted_table = self.tables["asserted_statements"]
        asserted_type_table = self.tables["type_statements"]
        literal_table = self.tables["literal_statements"]

        typetable = expression.alias(asserted_type_table, "typetable")
        quoted = expression.alias(quoted_table, "quoted")
        asserted = expression.alias(asserted_table, "asserted")
        literal = expression.alias(literal_table, "literal")

        if triple is not None:
            subject, predicate, obj = triple
            if predicate == RDF.type:
                # Select from asserted rdf:type partition and quoted table
                # (if a context is specified)
                clause = self.build_clause(typetable, subject, RDF.type, obj, Any, True)
                selects = [(typetable, clause, ASSERTED_TYPE_PARTITION), ]

            elif (predicate is None or
                    (isinstance(predicate, REGEXTerm) and
                        predicate.compiledExpr.match(RDF.type))):
                # Select from quoted partition (if context is specified),
                # literal partition if (obj is Literal or None) and
                # asserted non rdf:type partition (if obj is URIRef
                # or None)
                clause = self.build_clause(typetable, subject, RDF.type, obj, Any, True)
                selects = [(typetable, clause, ASSERTED_TYPE_PARTITION), ]

                if (not self.STRONGLY_TYPED_TERMS or
                        isinstance(obj, Literal) or
                        obj is None or
                        (self.STRONGLY_TYPED_TERMS and isinstance(obj, REGEXTerm))):
                    clause = self.build_clause(literal, subject, predicate, obj)
                    selects.append(
                        (literal, clause, ASSERTED_LITERAL_PARTITION))
                if not isinstance(obj, Literal) \
                        and not (isinstance(obj, REGEXTerm) and self.STRONGLY_TYPED_TERMS) \
                        or obj is None:
                    clause = self.build_clause(asserted, subject, predicate, obj)
                    selects.append((asserted, clause, ASSERTED_NON_TYPE_PARTITION))

            elif predicate:
                # select from asserted non rdf:type partition (optionally),
                # quoted partition (if context is specified), and literal
                # partition (optionally)
                selects = []
                if (not self.STRONGLY_TYPED_TERMS or
                        isinstance(obj, Literal) or obj is None or (
                            self.STRONGLY_TYPED_TERMS and isinstance(obj, REGEXTerm))):
                    clause = self.build_clause(literal, subject, predicate, obj)
                    selects.append(
                        (literal, clause, ASSERTED_LITERAL_PARTITION))
                if not isinstance(obj, Literal) \
                        and not (isinstance(obj, REGEXTerm) and self.STRONGLY_TYPED_TERMS) \
                        or obj is None:
                    clause = self.build_clause(asserted, subject, predicate, obj)
                    selects.append(
                        (asserted, clause, ASSERTED_NON_TYPE_PARTITION))

            clause = self.build_clause(quoted, subject, predicate, obj)
            selects.append((quoted, clause, QUOTED_PARTITION))
            q = union_select(selects, distinct=True, select_type=CONTEXT_SELECT)
        else:
            selects = [
                (typetable, None, ASSERTED_TYPE_PARTITION),
                (quoted, None, QUOTED_PARTITION),
                (asserted, None, ASSERTED_NON_TYPE_PARTITION),
                (literal, None, ASSERTED_LITERAL_PARTITION), ]
            q = union_select(selects, distinct=True, select_type=CONTEXT_SELECT)

        with self.engine.connect() as connection:
            res = connection.execute(q)
            rt = res.fetchall()
        for context in [rtTuple[0] for rtTuple in rt]:
            yield URIRef(context)

    # Namespace persistence interface implementation

    def bind(self, prefix, namespace):
        """Bind prefix for namespace."""
        with self.engine.begin() as connection:
            try:
                binds_table = self.tables["namespace_binds"]
                prefix = text_type(prefix)
                namespace = text_type(namespace)
                connection.execute(binds_table.delete().where(binds_table.c.prefix == prefix))
                connection.execute(binds_table.insert().values(prefix=prefix, uri=namespace))
            except Exception:
                _logger.exception("Namespace binding failed.")
                raise

    def prefix(self, namespace):
        """Prefix."""
        with self.engine.connect() as connection:
            nb_table = self.tables["namespace_binds"]
            namespace = text_type(namespace)
            s = select([nb_table.c.prefix]).where(nb_table.c.uri == namespace)
            res = connection.execute(s)
            rt = [rtTuple[0] for rtTuple in res.fetchall()]
            res.close()
            if rt and (rt[0] or rt[0] == ""):
                return rt[0]
        return None

    def namespace(self, prefix):
        res = None
        prefix_val = text_type(prefix)
        try:
            with self.engine.connect() as connection:
                nb_table = self.tables["namespace_binds"]
                s = select([nb_table.c.uri]).where(nb_table.c.prefix == prefix_val)
                res = connection.execute(s)
                rt = [rtTuple[0] for rtTuple in res.fetchall()]
                res.close()
                return rt and URIRef(rt[0]) or None
        except Exception:
            _logger.warning('exception in namespace retrieval', exc_info=True)
            return None

    def namespaces(self):
        with self.engine.connect() as connection:
            res = connection.execute(self.tables["namespace_binds"].select(distinct=True))
            for prefix, uri in res.fetchall():
                yield prefix, uri

    # Private methods

    def _create_table_definitions(self):
        self.metadata = MetaData()
        self.tables = {
            "asserted_statements": create_asserted_statements_table(self._interned_id, self.metadata),
            "type_statements": create_type_statements_table(self._interned_id, self.metadata),
            "literal_statements": create_literal_statements_table(self._interned_id, self.metadata),
            "quoted_statements": create_quoted_statements_table(self._interned_id, self.metadata),
            "namespace_binds": create_namespace_binds_table(self._interned_id, self.metadata),
        }

    def _get_build_command(self, triple, context=None, quoted=False):
        """
        Assemble the SQL Query text for adding an RDF triple to store.

        :param triple {tuple} - tuple of (subject, predicate, object) objects to add
        :param context - a `rdflib.URIRef` identifier for the graph namespace
        :param quoted {bool} - whether should treat as a quoted statement

        :returns {tuple} of (command_type, add_command, params):
            command_type: which kind of statement it is: literal, type, other
            statement: the literal SQL statement to execute (with unbound variables)
            params: the parameters for the SQL statement (e.g the variables to bind)

        """
        subject, predicate, obj = triple
        command_type = None
        if quoted or predicate != RDF.type:
            # Quoted statement or non rdf:type predicate
            # check if object is a literal
            if isinstance(obj, Literal):
                statement, params = self._build_literal_triple_sql_command(
                    subject,
                    predicate,
                    obj,
                    context,
                )
                command_type = "literal"
            else:
                statement, params = self._build_triple_sql_command(
                    subject,
                    predicate,
                    obj,
                    context,
                    quoted,
                )
                command_type = "other"
        elif predicate == RDF.type:
            # asserted rdf:type statement
            statement, params = self._build_type_sql_command(
                subject,
                obj,
                context,
            )
            command_type = "type"
        return command_type, statement, params

    def _remove_context(self, context):
        """Remove context."""
        assert context is not None
        quoted_table = self.tables["quoted_statements"]
        asserted_table = self.tables["asserted_statements"]
        asserted_type_table = self.tables["type_statements"]
        literal_table = self.tables["literal_statements"]

        with self.engine.begin() as connection:
            try:
                for table in [quoted_table, asserted_table,
                              asserted_type_table, literal_table]:
                    clause = self.build_context_clause(context, table)
                    connection.execute(table.delete(clause))
            except Exception:
                _logger.exception("Context removal failed.")
                raise

    def _verify_store_exists(self):
        """
        Verify store (e.g. all tables) exist.
        """

        for table_name in self.table_names:
            inspector = inspect(self.engine)
            if not inspector.has_table(table_name):
                _logger.critical("create_all() - table %s is not known", table_name)
                # The database exists, but one of the tables doesn't exist
                return CORRUPTED_STORE

        return VALID_STORE



"""pg8000 test."""
import logging
import os
import unittest

from nose import SkipTest
try:
    import pg8000
    assert pg8000 is not None
except ImportError:
    raise SkipTest("pg8000 not installed, skipping PgSQL tests")

from . import context_case
from . import graph_case


if os.environ.get("DB") != "pgsql":
    raise SkipTest("PgSQL not under test")

_logger = logging.getLogger(__name__)

sqlalchemy_url = os.environ.get(
    "DBURI",
    "postgresql+pg8000://postgres@localhost/test")


class SQLAPgSQLGraphTestCase(graph_case.GraphTestCase):
    """Graph test case."""

    storetest = True
    storename = "SQLAlchemy"
    uri = sqlalchemy_url
    create = True

    def setUp(self):
        """Setup."""
        super(SQLAPgSQLGraphTestCase, self).setUp(
            uri=self.uri, storename=self.storename)

    def tearDown(self):
        """Teardown."""
        super(SQLAPgSQLGraphTestCase, self).tearDown(uri=self.uri)


class SQLAPgSQLContextTestCase(context_case.ContextTestCase):
    """Context test case."""

    storetest = True
    storename = "SQLAlchemy"
    uri = sqlalchemy_url
    create = True

    def setUp(self):
        """Setup."""
        super(SQLAPgSQLContextTestCase, self).setUp(
            uri=self.uri, storename=self.storename)

    def tearDown(self):
        """Teardown."""
        super(SQLAPgSQLContextTestCase, self).tearDown(uri=self.uri)

    def testLenInMultipleContexts(self):
        """Test lin in multiple contexts, known issue."""
        raise SkipTest("Known issue.")

# SQLAPgSQLGraphTestCase.storetest = True
# SQLAPgSQLContextTestCase.storetest = True


if __name__ == "__main__":
    unittest.main()



import unittest

from rdflib import plugin
from rdflib.graph import Graph, Store


class TestOpen(unittest.TestCase):
    def test_open_corrupted(self):
        store = plugin.get('SQLAlchemy', Store)(identifier='open_test')
        self.graph = Graph(store, identifier='open_test')
        with self.assertRaises(RuntimeError):
            self.graph.open("sqlite:///test/corrupted.sqlite", create=False)



# -*- coding: utf-8 -*-
import unittest

from rdflib import Graph, URIRef, Literal, plugin, RDF
from rdflib.parser import StringInputSource
from six import PY3
from rdflib.store import Store


class GraphTestCase(unittest.TestCase):
    storetest = True
    identifier = URIRef("rdflib_test")

    michel = URIRef(u"michel")
    tarek = URIRef(u"tarek")
    bob = URIRef(u"bob")
    likes = URIRef(u"likes")
    hates = URIRef(u"hates")
    pizza = URIRef(u"pizza")
    cheese = URIRef(u"cheese")

    namespace_dc = "http://purl.org/dc/elements/1.1/"
    namespace_dct = "http://purl.org/dc/terms/"
    namespace_saws = "http://purl.org/saws/ontology#"

    def setUp(self, uri="sqlite://", storename=None):
        store = plugin.get(storename, Store)(identifier=self.identifier)
        self.graph = Graph(store, identifier=self.identifier)
        self.graph.open(uri, create=True)

    def tearDown(self, uri="sqlite://"):
        self.graph.destroy(uri)
        self.graph.close()

    def addStuff(self):
        tarek = self.tarek
        michel = self.michel
        bob = self.bob
        likes = self.likes
        hates = self.hates
        pizza = self.pizza
        cheese = self.cheese

        self.graph.add((tarek, likes, pizza))
        self.graph.add((tarek, likes, cheese))
        self.graph.add((michel, likes, pizza))
        self.graph.add((michel, likes, cheese))
        self.graph.add((bob, likes, cheese))
        self.graph.add((bob, hates, pizza))
        self.graph.add((bob, hates, michel))  # gasp!
        self.graph.commit()

    def removeStuff(self):
        tarek = self.tarek
        michel = self.michel
        bob = self.bob
        likes = self.likes
        hates = self.hates
        pizza = self.pizza
        cheese = self.cheese

        self.graph.remove((tarek, likes, pizza))
        self.graph.remove((tarek, likes, cheese))
        self.graph.remove((michel, likes, pizza))
        self.graph.remove((michel, likes, cheese))
        self.graph.remove((bob, likes, cheese))
        self.graph.remove((bob, hates, pizza))
        self.graph.remove((bob, hates, michel))  # gasp!

    def testAdd(self):
        self.addStuff()

    def testRemove(self):
        self.addStuff()
        self.removeStuff()

    def testTriples(self):
        tarek = self.tarek
        michel = self.michel
        bob = self.bob
        likes = self.likes
        hates = self.hates
        pizza = self.pizza
        cheese = self.cheese
        asserte = self.assertEqual
        triples = self.graph.triples
        Any = None

        self.addStuff()

        # unbound subjects
        asserte(len(list(triples((Any, likes, pizza)))), 2)
        asserte(len(list(triples((Any, hates, pizza)))), 1)
        asserte(len(list(triples((Any, likes, cheese)))), 3)
        asserte(len(list(triples((Any, hates, cheese)))), 0)

        # unbound objects
        asserte(len(list(triples((michel, likes, Any)))), 2)
        asserte(len(list(triples((tarek, likes, Any)))), 2)
        asserte(len(list(triples((bob, hates, Any)))), 2)
        asserte(len(list(triples((bob, likes, Any)))), 1)

        # unbound predicates
        asserte(len(list(triples((michel, Any, cheese)))), 1)
        asserte(len(list(triples((tarek, Any, cheese)))), 1)
        asserte(len(list(triples((bob, Any, pizza)))), 1)
        asserte(len(list(triples((bob, Any, michel)))), 1)

        # unbound subject, objects
        asserte(len(list(triples((Any, hates, Any)))), 2)
        asserte(len(list(triples((Any, likes, Any)))), 5)

        # unbound predicates, objects
        asserte(len(list(triples((michel, Any, Any)))), 2)
        asserte(len(list(triples((bob, Any, Any)))), 3)
        asserte(len(list(triples((tarek, Any, Any)))), 2)

        # unbound subjects, predicates
        asserte(len(list(triples((Any, Any, pizza)))), 3)
        asserte(len(list(triples((Any, Any, cheese)))), 3)
        asserte(len(list(triples((Any, Any, michel)))), 1)

        # all unbound
        asserte(len(list(triples((Any, Any, Any)))), 7)
        self.removeStuff()
        asserte(len(list(triples((Any, Any, Any)))), 0)

    def testConnected(self):
        graph = self.graph
        self.addStuff()
        self.assertEqual(True, graph.connected())

        jeroen = URIRef("jeroen")
        unconnected = URIRef("unconnected")

        graph.add((jeroen, self.likes, unconnected))

        self.assertEqual(False, graph.connected())

    def testSub(self):
        g1 = Graph()
        g2 = Graph()

        tarek = self.tarek
        # michel = self.michel
        bob = self.bob
        likes = self.likes
        # hates = self.hates
        pizza = self.pizza
        cheese = self.cheese

        g1.add((tarek, likes, pizza))
        g1.add((bob, likes, cheese))

        g2.add((bob, likes, cheese))

        g3 = g1 - g2

        self.assertEqual(len(g3), 1)
        self.assertEqual((tarek, likes, pizza) in g3, True)
        self.assertEqual((tarek, likes, cheese) in g3, False)

        self.assertEqual((bob, likes, cheese) in g3, False)

        g1 -= g2

        self.assertEqual(len(g1), 1)
        self.assertEqual((tarek, likes, pizza) in g1, True)
        self.assertEqual((tarek, likes, cheese) in g1, False)

        self.assertEqual((bob, likes, cheese) in g1, False)

    def testGraphAdd(self):
        g1 = Graph()
        g2 = Graph()

        tarek = self.tarek
        # michel = self.michel
        bob = self.bob
        likes = self.likes
        # hates = self.hates
        pizza = self.pizza
        cheese = self.cheese

        g1.add((tarek, likes, pizza))

        g2.add((bob, likes, cheese))

        g3 = g1 + g2

        self.assertEqual(len(g3), 2)
        self.assertEqual((tarek, likes, pizza) in g3, True)
        self.assertEqual((tarek, likes, cheese) in g3, False)

        self.assertEqual((bob, likes, cheese) in g3, True)

        g1 += g2

        self.assertEqual(len(g1), 2)
        self.assertEqual((tarek, likes, pizza) in g1, True)
        self.assertEqual((tarek, likes, cheese) in g1, False)

        self.assertEqual((bob, likes, cheese) in g1, True)

    def testGraphIntersection(self):
        g1 = Graph()
        g2 = Graph()

        tarek = self.tarek
        michel = self.michel
        bob = self.bob
        likes = self.likes
        # hates = self.hates
        pizza = self.pizza
        cheese = self.cheese

        g1.add((tarek, likes, pizza))
        g1.add((michel, likes, cheese))

        g2.add((bob, likes, cheese))
        g2.add((michel, likes, cheese))

        g3 = g1 * g2

        self.assertEqual(len(g3), 1)
        self.assertEqual((tarek, likes, pizza) in g3, False)
        self.assertEqual((tarek, likes, cheese) in g3, False)

        self.assertEqual((bob, likes, cheese) in g3, False)

        self.assertEqual((michel, likes, cheese) in g3, True)

        g1 *= g2

        self.assertEqual(len(g1), 1)

        self.assertEqual((tarek, likes, pizza) in g1, False)
        self.assertEqual((tarek, likes, cheese) in g1, False)

        self.assertEqual((bob, likes, cheese) in g1, False)

        self.assertEqual((michel, likes, cheese) in g1, True)

    def testStoreLiterals(self):
        bob = self.bob
        says = URIRef(u"says")
        hello = Literal(u"hello", lang="en")
        konichiwa = Literal(u"こんにちは", lang="ja")
        something = Literal(u"something")

        self.graph.add((bob, says, hello))
        self.graph.add((bob, says, konichiwa))
        self.graph.add((bob, says, something))
        self.graph.commit()

        objs = list(self.graph.objects(subject=bob, predicate=says))
        for o in objs:
            if o.value == u"hello":
                self.assertEqual(o.language, "en")
            elif o.value == u"こんにちは":
                self.assertEqual(o.language, "ja")
            elif o.value == u"something":
                self.assertEqual(o.language, None)
            else:
                self.fail()
        self.assertEqual(len(list(objs)), 3)

    def testStoreLiteralsXml(self):
        bob = self.bob
        says = URIRef(u"http://www.rdflib.net/terms/says")
        objects = [
            Literal(u"I'm the one", lang="en"),
            Literal(u"こんにちは", lang="ja"),
            Literal(u"les garçons à Noël reçoivent des œufs", lang="fr")]

        testdoc = (PY3 and bytes(xmltestdocXml, "UTF-8")) or xmltestdocXml

        self.graph.parse(StringInputSource(testdoc), format="xml")

        objs = list(self.graph)
        self.assertEqual(len(objs), 3)
        for o in objs:
            self.assertEqual(o[0], bob)
            self.assertEqual(o[1], says)
            self.assertIn(o[2], objects)

    def testStoreLiteralsXmlQuote(self):
        bob = self.bob
        says = URIRef(u"http://www.rdflib.net/terms/says")
        imtheone = Literal(u"I'm the one", lang="en")

        testdoc = (PY3 and bytes(xmltestdocXmlQuote, "UTF-8")) or xmltestdocXmlQuote

        self.graph.parse(StringInputSource(testdoc), format="xml")

        objs = list(self.graph)
        self.assertEqual(len(objs), 1)
        o = objs[0]
        self.assertEqual(o, (bob, says, imtheone))

    def testBindNamespace(self):
        """ Check that bound namespaced with prefix (including empty ones) are correctly kept """
        self.graph.bind("", self.namespace_dc)
        self.graph.bind("dct", self.namespace_dct)
        self.assertEqual(
            self.graph.qname(self.namespace_dct + u"title"), "dct:title",
            "Prefixed namespace should be stored and retrieved"
        )
        self.assertEqual(
            self.graph.qname(self.namespace_dc + u"title"), "title",
            "Empty prefixes for namespace should be stored and retrieved"
        )
        self.assertEqual(
            self.graph.qname(self.namespace_saws + u"title"), "ns1:title",
            "Unknown prefixes for namespace should be transformed to nsX"
        )

    def testTriplesChoices(self):
        likes = self.likes
        pizza = self.pizza
        cheese = self.cheese
        tarek = self.tarek
        michel = self.michel
        bob = self.bob
        self.addStuff()
        trips = self.graph.triples_choices((None, likes, [pizza, cheese]))
        self.assertEqual(
            set(trips),
            set([(tarek, likes, pizza),
                 (tarek, likes, pizza),
                 (tarek, likes, cheese),
                 (michel, likes, pizza),
                 (michel, likes, cheese),
                 (bob, likes, cheese)])
        )

    def test_type_add(self):
        trip = (URIRef('http://example.org#type-add'), RDF.type, URIRef('http://example.org/cra'))
        self.graph.add(trip)
        self.graph.add(trip)
        # No exception raised

    def test_type_addn(self):
        quad = (URIRef('http://example.org#type-addn'), RDF.type, URIRef('http://example.org/cra'), self.graph)
        self.graph.addN([quad, quad])
        # No exception raised

    def test_add(self):
        trip = (URIRef('http://example.org#add'), URIRef('http://example.org/blah'), URIRef('http://example.org/cra'))
        self.graph.add(trip)
        self.graph.add(trip)
        # No exception raised

    def test_addn(self):
        quad = (URIRef('http://example.org#addn'),
                URIRef('http://example.org/blah'),
                URIRef('http://example.org/cra'),
                self.graph)
        self.graph.addN([quad, quad])
        # No exception raised

    def test_namespace_change_prefix_binding(self):
        nm = self.graph.namespace_manager
        nm.bind('change_binding', URIRef('http://example.org/change-binding-1#'),
                replace=True)
        nm.bind('change_binding', URIRef('http://example.org/change-binding-2#'),
                replace=True)
        self.assertIn(
            ('change_binding', URIRef('http://example.org/change-binding-2#')),
            list(nm.namespaces())
        )

    def test_namespace_rebind_prefix(self):
        nm = self.graph.namespace_manager
        nm.bind('rebind', URIRef('http://example.org/rebind#'))
        nm.bind('rebind', URIRef('http://example.org/rebind#'))

    def test_add_duplicate_length(self):
        '''
        Test that adding duplicate triples doesn't increase the length of the graph
        '''
        trip = (URIRef('http://example.org#add'), URIRef('http://example.org/blah'), URIRef('http://example.org/cra'))
        self.graph.add(trip)
        self.graph.add(trip)
        self.assertEqual(len(self.graph), 1)
    # additional tests
    # - add "duplicate" triples and query -- ensure the graph length counts only distinct
    #   triples
    # - add duplicate triples and query -- ensure there are no duplicate entries in the
    #   query result
    # - repeat the above two for type triples
    # - test with quoted graphs (not even sure how that works)


xmltestdoc = """<?xml version="1.0" encoding="UTF-8"?>
<rdf:RDF
   xmlns="http://example.org/"
   xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
>
  <rdf:Description rdf:about="http://example.org/a">
    <b rdf:resource="http://example.org/c"/>
  </rdf:Description>
</rdf:RDF>
"""

xmltestdocXml = """<?xml version="1.0" encoding="UTF-8"?>
<rdf:RDF
   xmlns:ns1="http://www.rdflib.net/terms/"
   xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
>
  <rdf:Description rdf:about="bob">
    <ns1:says xml:lang="en">I'm the one</ns1:says>
  </rdf:Description>
  <rdf:Description rdf:about="bob">
    <ns1:says xml:lang="ja">こんにちは</ns1:says>
  </rdf:Description>
  <rdf:Description rdf:about="bob">
    <ns1:says xml:lang="fr">les garçons à Noël reçoivent des œufs</ns1:says>
  </rdf:Description>
</rdf:RDF>
"""

xmltestdocXmlQuote = """<?xml version="1.0" encoding="UTF-8"?>
<rdf:RDF
   xmlns:ns1="http://www.rdflib.net/terms/"
   xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
>
  <rdf:Description rdf:about="bob">
    <ns1:says xml:lang="en">I'm the one</ns1:says>
  </rdf:Description>
</rdf:RDF>
"""


n3testdoc = """@prefix : <http://example.org/> .

:a :b :c .
"""

nttestdoc = "<http://example.org/a> <http://example.org/b> <http://example.org/c> .\n"

if __name__ == "__main__":
    unittest.main()



from __future__ import print_function
import unittest
import gc
import os
import shutil
from time import time
from tempfile import mkdtemp

from rdflib import Graph
from six.moves.urllib.request import pathname2url

try:
    from rdflib.plugins.stores.memory import Memory
except ImportError:
    # rdflib<6.0.0
    from rdflib.plugins.memory import IOMemory as Memory


class StoreTestCase(unittest.TestCase):

    """
    Test case for testing store performance... probably should be
    something other than a unit test... but for now we'll add it as a
    unit test.
    """
    store = None
    path = None
    storetest = True
    performancetest = True

    def setUp(self):
        self.gcold = gc.isenabled()
        gc.collect()
        gc.disable()
        if self.store is None:
            store = Memory()
        else:
            store = self.store
        self.graph = Graph(store=store)
        self.tempdir = None
        if not self.path:
            self.tempdir = mkdtemp()
            path = pathname2url(self.tempdir)
        else:
            path = self.path
        self.path = path
        self.graph.open(self.path, create=True)
        self.input = Graph()

    def tearDown(self):
        self.graph.close()
        if self.gcold:
            gc.enable()
        self.graph.close()
        del self.graph
        tempdir = getattr(self, 'tempdir', None)
        if tempdir is not None:
            shutil.rmtree(tempdir)

    def testTime(self):
        # number = 1
        print('"Load %s": [' % self.store)
        for i in ['500triples', '1ktriples', '2ktriples',
                  '3ktriples', '5ktriples', '10ktriples',
                  '25ktriples']:
            inputloc = os.getcwd() + '/test/sp2b/%s.n3' % i
            # cleanup graph's so that BNodes in input data
            # won't create random results
            self.input = Graph()
            self.graph.remove((None, None, None))
            res = self._testInput(inputloc)
            print("Loaded %5d triples in %ss" % (len(self.graph), res.strip()))
        print("],")
        print('"Read %s": [' % self.store)
        t0 = time()
        for _i in self.graph.triples((None, None, None)):
            pass
        self.assertEqual(len(self.graph), 25161)
        t1 = time()
        print("%.3gs" % (t1 - t0))
        print("],")
        print('"Delete %s": [' % self.store)
        t0 = time()
        self.graph.remove((None, None, None))
        self.assertEqual(len(self.graph), 0)
        t1 = time()
        print("%.3g " % (t1 - t0))
        print("],")

    def _testInput(self, inputloc):
        store = self.graph
        self.input.parse(location=inputloc, format="n3")

        t0 = time()
        store.addN(tuple(t) + (store,) for t in self.input)
        t1 = time()
        return "%.3g " % (t1 - t0)


class SQLAlchemyStoreTestCase(StoreTestCase):
    """SQLAlchemy Store."""

    store = "SQLAlchemy"

    def setUp(self):
        """Setup."""
        self.store = "SQLAlchemy"
        self.path = "sqlite:///%(here)s/test/tmpdb.sqlite" % {
            "here": os.getcwd()
        }
        StoreTestCase.setUp(self)


if __name__ == '__main__':
    unittest.main()



from rdflib.namespace import RDF
from six import text_type
from sqlalchemy.sql import expression, functions

from rdflib_sqlalchemy.constants import (
    ASSERTED_TYPE_PARTITION,
    ASSERTED_NON_TYPE_PARTITION,
    ASSERTED_LITERAL_PARTITION,
    QUOTED_PARTITION,
    CONTEXT_SELECT,
    COUNT_SELECT,
    FULL_TRIPLE_PARTITIONS,
    TRIPLE_SELECT,
)


def query_analysis(query, store, connection):
    """
    Helper function.

    For executing EXPLAIN on all dispatched SQL statements -
    for the pupose of analyzing index usage.

    """
    res = connection.execute("explain " + query)
    rt = res.fetchall()[0]
    table, joinType, posKeys, _key, key_len, \
        comparedCol, rowsExamined, extra = rt
    if not _key:
        assert joinType == "ALL"
        if not hasattr(store, "queryOptMarks"):
            store.queryOptMarks = {}
        hits = store.queryOptMarks.get(("FULL SCAN", table), 0)
        store.queryOptMarks[("FULL SCAN", table)] = hits + 1

    if not hasattr(store, "queryOptMarks"):
        store.queryOptMarks = {}
    hits = store.queryOptMarks.get((_key, table), 0)
    store.queryOptMarks[(_key, table)] = hits + 1


def union_select(select_components, distinct=False, select_type=TRIPLE_SELECT):
    """
    Helper function for building union all select statement.

    Args:
        select_components (iterable of tuples): Indicates the table and table type
            (table_name, where_clause_string, table_type)
        distinct (bool): Whether to eliminate duplicate results
        select_type (int): From `rdflib_sqlalchemy.constants`. Either `COUNT_SELECT`,
            `CONTEXT_SELECT`, `TRIPLE_SELECT`

    """
    selects = []
    for table, whereClause, tableType in select_components:

        if select_type == COUNT_SELECT:
            c = table.c
            if tableType == ASSERTED_TYPE_PARTITION:
                cols = [c.member, c.klass]
            elif tableType in (ASSERTED_LITERAL_PARTITION, ASSERTED_NON_TYPE_PARTITION, QUOTED_PARTITION):
                cols = [c.subject, c.predicate, c.object]
            else:
                raise ValueError('Unrecognized table type {}'.format(tableType))
            select_clause = expression.select([functions.count().label('aCount')]).select_from(
                expression.select(cols, whereClause).distinct().select_from(table))
        elif select_type == CONTEXT_SELECT:
            select_clause = expression.select([table.c.context], whereClause)
        elif tableType in FULL_TRIPLE_PARTITIONS:
            select_clause = table.select(whereClause)
        elif tableType == ASSERTED_TYPE_PARTITION:
            select_clause = expression.select(
                [table.c.id.label("id"),
                 table.c.member.label("subject"),
                 expression.literal(text_type(RDF.type)).label("predicate"),
                 table.c.klass.label("object"),
                 table.c.context.label("context"),
                 table.c.termComb.label("termcomb"),
                 expression.literal_column("NULL").label("objlanguage"),
                 expression.literal_column("NULL").label("objdatatype")],
                whereClause)
        elif tableType == ASSERTED_NON_TYPE_PARTITION:
            select_clause = expression.select(
                [c for c in table.columns] +
                [expression.literal_column("NULL").label("objlanguage"),
                 expression.literal_column("NULL").label("objdatatype")],
                whereClause,
                from_obj=[table])

        selects.append(select_clause)

    order_statement = []
    if select_type == TRIPLE_SELECT:
        order_statement = [
            expression.literal_column("subject"),
            expression.literal_column("predicate"),
            expression.literal_column("object"),
        ]
    if distinct and select_type != COUNT_SELECT:
        return expression.union(*selects, **{"order_by": order_statement})
    else:
        return expression.union_all(*selects, **{"order_by": order_statement})



"""Base classes for the store."""
from rdflib import Literal
from rdflib.graph import Graph, QuotedGraph
from rdflib.plugins.stores.regexmatching import REGEXTerm
from sqlalchemy.sql import expression

from rdflib_sqlalchemy.termutils import (
    type_to_term_combination,
    statement_to_term_combination,
)


class SQLGeneratorMixin(object):
    """SQL statement generator mixin for the SQLAlchemy store."""

    def _build_type_sql_command(self, member, klass, context):
        """Build an insert command for a type table."""
        # columns: member,klass,context
        rt = self.tables["type_statements"].insert()
        return rt, {
            "member": member,
            "klass": klass,
            "context": context.identifier,
            "termComb": int(type_to_term_combination(member, klass, context))
        }

    def _build_literal_triple_sql_command(self, subject, predicate, obj, context):
        """
        Build an insert command for literal triples.

        These triples correspond to RDF statements where the object is a Literal,
        e.g. `rdflib.Literal`.

        """
        triple_pattern = int(
            statement_to_term_combination(subject, predicate, obj, context)
        )

        command = self.tables["literal_statements"].insert()
        values = {
            "subject": subject,
            "predicate": predicate,
            "object": obj,
            "context": context.identifier,
            "termComb": triple_pattern,
            "objLanguage": isinstance(obj, Literal) and obj.language or None,
            "objDatatype": isinstance(obj, Literal) and obj.datatype or None,
        }
        return command, values

    def _build_triple_sql_command(self, subject, predicate, obj, context, quoted):
        """
        Build an insert command for regular triple table.

        """
        stmt_table = (quoted and
                      self.tables["quoted_statements"] or
                      self.tables["asserted_statements"])

        triple_pattern = statement_to_term_combination(
            subject,
            predicate,
            obj,
            context,
        )
        command = stmt_table.insert()

        if quoted:
            params = {
                "subject": subject,
                "predicate": predicate,
                "object": obj,
                "context": context.identifier,
                "termComb": triple_pattern,
                "objLanguage": isinstance(obj, Literal) and obj.language or None,
                "objDatatype": isinstance(obj, Literal) and obj.datatype or None
            }
        else:
            params = {
                "subject": subject,
                "predicate": predicate,
                "object": obj,
                "context": context.identifier,
                "termComb": triple_pattern,
            }
        return command, params

    def build_clause(self, table, subject, predicate, obj, context=None, typeTable=False):
        """Build WHERE clauses for the supplied terms and, context."""
        if typeTable:
            clauseList = [
                self.build_type_member_clause(subject, table),
                self.build_type_class_clause(obj, table),
                self.build_context_clause(context, table)
            ]
        else:
            clauseList = [
                self.build_subject_clause(subject, table),
                self.build_predicate_clause(predicate, table),
                self.build_object_clause(obj, table),
                self.build_context_clause(context, table),
                self.build_literal_datatype_clause(obj, table),
                self.build_literal_language_clause(obj, table)
            ]

        clauseList = [clause for clause in clauseList if clause is not None]
        if clauseList:
            return expression.and_(*clauseList)
        else:
            return None

    def build_literal_datatype_clause(self, obj, table):
        """Build Literal and datatype clause."""
        if isinstance(obj, Literal) and obj.datatype is not None:
            return table.c.objDatatype == obj.datatype
        else:
            return None

    def build_literal_language_clause(self, obj, table):
        """Build Literal and language clause."""
        if isinstance(obj, Literal) and obj.language is not None:
            return table.c.objLanguage == obj.language
        else:
            return None

    # Where Clause  utility Functions
    # The predicate and object clause builders are modified in order
    # to optimize subjects and objects utility functions which can
    # take lists as their last argument (object, predicate - respectively)

    def build_subject_clause(self, subject, table):
        """Build Subject clause."""
        if isinstance(subject, REGEXTerm):
            # TODO: this work only in mysql. Must adapt for postgres and sqlite
            return table.c.subject.op("REGEXP")(subject)
        elif isinstance(subject, list):
            # clauseStrings = [] --- unused
            return expression.or_(
                *[self.build_subject_clause(s, table) for s in subject if s])
        elif isinstance(subject, (QuotedGraph, Graph)):
            return table.c.subject == subject.identifier
        elif subject is not None:
            return table.c.subject == subject
        else:
            return None

    def build_predicate_clause(self, predicate, table):
        """
        Build Predicate clause.

        Capable of taking a list of predicates as well, in which case
        subclauses are joined with 'OR'.

        """
        if isinstance(predicate, REGEXTerm):
            # TODO: this work only in mysql. Must adapt for postgres and sqlite
            return table.c.predicate.op("REGEXP")(predicate)
        elif isinstance(predicate, list):
            return expression.or_(
                *[self.build_predicate_clause(p, table) for p in predicate if p])
        elif predicate is not None:
            return table.c.predicate == predicate
        else:
            return None

    def build_object_clause(self, obj, table):
        """
        Build Object clause.

        Capable of taking a list of objects as well, in which case subclauses
        are joined with 'OR'.

        """
        if isinstance(obj, REGEXTerm):
            # TODO: this work only in mysql. Must adapt for postgres and sqlite
            return table.c.object.op("REGEXP")(obj)
        elif isinstance(obj, list):
            return expression.or_(
                *[self.build_object_clause(o, table) for o in obj if o])
        elif isinstance(obj, (QuotedGraph, Graph)):
            return table.c.object == obj.identifier
        elif obj is not None:
            return table.c.object == obj
        else:
            return None

    def build_context_clause(self, context, table):
        """Build Context clause."""
        if isinstance(context, REGEXTerm):
            # TODO: this work only in mysql. Must adapt for postgres and sqlite
            return table.c.context.op("regexp")(context.identifier)
        elif context is not None and context.identifier is not None:
            return table.c.context == context.identifier
        else:
            return None

    def build_type_member_clause(self, subject, table):
        """Build Type Member clause."""
        if isinstance(subject, REGEXTerm):
            # TODO: this work only in mysql. Must adapt for postgres and sqlite
            return table.c.member.op("regexp")(subject)
        elif isinstance(subject, list):
            return expression.or_(
                *[self.build_type_member_clause(s, table) for s in subject if s])
        elif subject is not None:
            return table.c.member == subject
        else:
            return None

    def build_type_class_clause(self, obj, table):
        """Build Type Class clause."""
        if isinstance(obj, REGEXTerm):
            # TODO: this work only in mysql. Must adapt for postgres and sqlite
            return table.c.klass.op("regexp")(obj)
        elif isinstance(obj, list):
            return expression.or_(
                *[self.build_type_class_clause(o, table) for o in obj if o])
        elif obj is not None:
            return obj and table.c.klass == obj
        else:
            return None



"""Constant definitions"""
from rdflib import BNode, Literal, URIRef, Variable


COUNT_SELECT = 0
''' Return the count of the number of result tuples '''

CONTEXT_SELECT = 1
''' Return the matching contexts '''

TRIPLE_SELECT = 2
''' Return the matching triples '''

TRIPLE_SELECT_NO_ORDER = 3
''' Return the matching triples in any order '''

ASSERTED_NON_TYPE_PARTITION = 3
''' Table of non-type triples '''

ASSERTED_TYPE_PARTITION = 4
''' Table of type triples '''

QUOTED_PARTITION = 5
''' Table of `quoted`_ type triples

.. _quoted: https://rdflib.readthedocs.io/en/stable/univrdfstore.html#terminology
'''

ASSERTED_LITERAL_PARTITION = 6
''' Table of triples with literals '''

FULL_TRIPLE_PARTITIONS = [QUOTED_PARTITION, ASSERTED_LITERAL_PARTITION]

INTERNED_PREFIX = "kb_"

TERM_COMBINATIONS = dict([(term, index) for index, term in enumerate([
    "UUUU", "UUUB", "UUUF", "UUVU", "UUVB", "UUVF", "UUBU", "UUBB", "UUBF",
    "UULU", "UULB", "UULF", "UUFU", "UUFB", "UUFF",
    #
    "UVUU", "UVUB", "UVUF", "UVVU", "UVVB", "UVVF", "UVBU", "UVBB", "UVBF",
    "UVLU", "UVLB", "UVLF", "UVFU", "UVFB", "UVFF",
    #
    "VUUU", "VUUB", "VUUF", "VUVU", "VUVB", "VUVF", "VUBU", "VUBB", "VUBF",
    "VULU", "VULB", "VULF", "VUFU", "VUFB", "VUFF",
    #
    "VVUU", "VVUB", "VVUF", "VVVU", "VVVB", "VVVF", "VVBU", "VVBB", "VVBF",
    "VVLU", "VVLB", "VVLF", "VVFU", "VVFB", "VVFF",
    #
    "BUUU", "BUUB", "BUUF", "BUVU", "BUVB", "BUVF", "BUBU", "BUBB", "BUBF",
    "BULU", "BULB", "BULF", "BUFU", "BUFB", "BUFF",
    #
    "BVUU", "BVUB", "BVUF", "BVVU", "BVVB", "BVVF", "BVBU", "BVBB", "BVBF",
    "BVLU", "BVLB", "BVLF", "BVFU", "BVFB", "BVFF",
    #
    "FUUU", "FUUB", "FUUF", "FUVU", "FUVB", "FUVF", "FUBU", "FUBB", "FUBF",
    "FULU", "FULB", "FULF", "FUFU", "FUFB", "FUFF",
    #
    "FVUU", "FVUB", "FVUF", "FVVU", "FVVB", "FVVF", "FVBU", "FVBB", "FVBF",
    "FVLU", "FVLB", "FVLF", "FVFU", "FVFB", "FVFF",
])])

REVERSE_TERM_COMBINATIONS = dict([
    (value, key)
    for key, value in TERM_COMBINATIONS.items()
])

TERM_INSTANTIATION_DICT = {
    "U": URIRef,
    "B": BNode,
    "V": Variable,
    "L": Literal
}



#!/usr/bin/env python
from setuptools import setup

project = "rdflib-sqlalchemy"
version = "0.4.1.dev20211114152243"


setup(
    name=project,
    version=version,
    description="rdflib extension adding SQLAlchemy as an AbstractSQLStore back-end store",
    author="Graham Higgins, Adam Ever-Hadani",
    author_email="gjhiggins@gmail.com, adamhadani@globality.com",
    url="http://github.com/RDFLib/rdflib-sqlalchemy",
    packages=["rdflib_sqlalchemy"],
    download_url="https://github.com/RDFLib/rdflib-sqlalchemy/zipball/master",
    license="BSD",
    platforms=["any"],
    long_description="""
    SQLAlchemy store formula-aware implementation.
    It stores its triples in the following partitions:

    * Asserted non rdf:type statements
    * Asserted rdf:type statements (in a table which models Class membership).
      The motivation for this partition is primarily improved query speed and
      scalability as most graphs will always have more rdf:type statements than
      others.
    * All Quoted statements

    In addition it persists namespace mappings in a separate table
    """,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: BSD License",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Operating System :: OS Independent",
        "Natural Language :: English",
    ],
    install_requires=[
        "alembic>=0.8.8",
        "rdflib>=4.0",
        "six>=1.10.0",
        "SQLAlchemy>=1.1.4",
    ],
    test_suite="nose.collector",
    entry_points={
        'rdf.plugins.store': [
            'SQLAlchemy = rdflib_sqlalchemy.store:SQLAlchemy'
        ]
    }
)



"""Statistical summary of store statements mixin"""
from sqlalchemy.orm.session import Session
from sqlalchemy.sql import func


def get_group_by_count(session, group_by_column):
    """
    Construct SQL query to get counts for distinct values using GROUP BY.

    :param session - sqlalchemy `Session` instance
    :param group_by_column sqlalchemy `Column` instance
    :returns {dict} dictionary mapping from value to count

    """
    return dict(
        session.query(
            group_by_column,
            func.count(group_by_column)
        ).group_by(group_by_column).all()
    )


class StatisticsMixin(object):

    def statistics(self, asserted_statements=True, literals=True, types=True):
        """Store statistics."""
        statistics = {
            "store": dict(total_num_statements=len(self)),
        }

        with self.engine.connect() as connection:
            session = Session(bind=connection)
            if asserted_statements:
                table = self.tables["asserted_statements"]
                group_by_column = table.c.predicate
                statistics["asserted_statements"] = get_group_by_count(session, group_by_column)
            if literals:
                table = self.tables["literal_statements"]
                group_by_column = table.c.predicate
                statistics["literals"] = get_group_by_count(session, group_by_column)
            if types:
                table = self.tables["type_statements"]
                group_by_column = table.c.klass
                statistics["types"] = get_group_by_count(session, group_by_column)

        return statistics



import unittest

try:
    from unittest.mock import patch, MagicMock
except ImportError:
    from mock import patch, MagicMock

import six

from rdflib import (
    ConjunctiveGraph,
    Literal,
    URIRef,
    plugin
)
from rdflib.store import Store

from rdflib_sqlalchemy import registerplugins
from sqlalchemy.sql.selectable import Select


michel = URIRef(u"michel")
likes = URIRef(u"likes")
pizza = URIRef(u"pizza")


class mock_cursor():
    def execute(x):
        raise Exception("Forced exception")


class ConfigTest(unittest.TestCase):
    '''
    Test configuration with a dict
    '''

    def setUp(self):
        self.store = plugin.get("SQLAlchemy", Store)()
        self.graph = ConjunctiveGraph(self.store)

    def tearDown(self):
        self.graph.close()

    def test_success(self):
        with patch('rdflib_sqlalchemy.store.sqlalchemy') as p:
            self.graph.open({'url': 'sqlite://', 'random_key': 'something'}, create=True)
            p.create_engine.assert_called_with('sqlite://', random_key='something')

    def test_no_url(self):
        with patch('rdflib_sqlalchemy.store.sqlalchemy'):
            with six.assertRaisesRegex(self, Exception, '.*url.*'):
                self.graph.open({'random_key': 'something'}, create=True)


class SQLATestCase(unittest.TestCase):
    identifier = URIRef("rdflib_test")
    dburi = Literal("sqlite://")

    def setUp(self):
        self.store = plugin.get(
            "SQLAlchemy", Store)(identifier=self.identifier, configuration=self.dburi)
        self.graph = ConjunctiveGraph(self.store, identifier=self.identifier)
        self.graph.open(self.dburi, create=True)

    def tearDown(self):
        self.graph.destroy(self.dburi)
        self.graph.close()

    def test_registerplugins(self):
        # I doubt this is quite right for a fresh pip installation,
        # this test is mainly here to fill a coverage gap.
        registerplugins()
        self.assertIsNotNone(plugin.get("SQLAlchemy", Store))
        p = plugin._plugins
        self.assertIn(("SQLAlchemy", Store), p)
        del p[("SQLAlchemy", Store)]
        plugin._plugins = p
        registerplugins()
        self.assertIn(("SQLAlchemy", Store), p)

    def test_namespaces(self):
        self.assertNotEqual(list(self.graph.namespaces()), [])

    def test_contexts_without_triple(self):
        self.assertEqual(list(self.graph.contexts()), [])

    def test_contexts_result(self):
        ctx_id = URIRef('http://example.org/context')
        g = self.graph.get_context(ctx_id)
        g.add((michel, likes, pizza))
        actual = list(self.store.contexts())
        self.assertEqual(actual[0], ctx_id)

    def test_contexts_with_triple(self):
        statemnt = (michel, likes, pizza)
        self.assertEqual(list(self.graph.contexts(triple=statemnt)), [])

    def test__len(self):
        self.assertEqual(self.store.__len__(), 0)

    def test__remove_context(self):
        ctx_id = URIRef('http://example.org/context')
        g = self.graph.get_context(ctx_id)
        g.add((michel, likes, pizza))
        self.store._remove_context(g)
        self.assertEqual(list(self.store.contexts()), [])

    def test_triples_choices(self):
        # Create a mock for the sqlalchemy engine so we can capture the arguments
        p = MagicMock(name='engine')
        self.store.engine = p

        # Set this so we're not including selects for both asserted and literal tables for
        # a choice
        self.store.STRONGLY_TYPED_TERMS = True
        # Set the grouping of terms
        self.store.max_terms_per_where = 2
        # force execution of the generator
        for x in self.store.triples_choices((None, likes, [michel, pizza, likes])):
            pass
        args = p.connect().__enter__().execute.call_args[0]
        children = args[0].get_children(column_collections=False)
        # Expect two selects: one for the first two choices plus one for the last one
        self.assertEqual(sum(1 for c in children if isinstance(c, Select)), 2)


if __name__ == "__main__":
    unittest.main()



import unittest

from rdflib import Literal, SDO, Graph

from rdflib_sqlalchemy.termutils import type_to_term_combination


class TypeToTermCombinationTestCase(unittest.TestCase):
    """Test the type_to_term_combination function."""

    def test_luuu(self):
        """Literal is not a valid subject for a triple."""
        with self.assertRaises(ValueError):
            type_to_term_combination(
                member=Literal('https://example.org'),
                klass=SDO.WebSite,
                context=Graph(identifier='local://test-graph/'),
            )
